/*******************************************************************************
 *     ___                  _   ____  ____
 *    / _ \ _   _  ___  ___| |_|  _ \| __ )
 *   | | | | | | |/ _ \/ __| __| | | |  _ \
 *   | |_| | |_| |  __/\__ \ |_| |_| | |_) |
 *    \__\_\\__,_|\___||___/\__|____/|____/
 *
 *  Copyright (c) 2014-2019 Appsicle
 *  Copyright (c) 2019-2022 QuestDB
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 ******************************************************************************/

package io.questdb;

import io.questdb.cairo.*;
import io.questdb.cutlass.http.HttpServer;
import io.questdb.cutlass.json.JsonException;
import io.questdb.cutlass.line.tcp.LineTcpReceiver;
import io.questdb.cutlass.line.udp.LineUdpReceiver;
import io.questdb.cutlass.line.udp.LinuxMMLineUdpReceiver;
import io.questdb.cutlass.pgwire.PGWireServer;
import io.questdb.griffin.FunctionFactory;
import io.questdb.griffin.FunctionFactoryCache;
import io.questdb.jit.JitUtil;
import io.questdb.log.Log;
import io.questdb.log.LogFactory;
import io.questdb.log.LogRecord;
import io.questdb.mp.WorkerPool;
import io.questdb.network.NetworkError;
import io.questdb.std.*;
import io.questdb.std.datetime.millitime.Dates;
import io.questdb.std.str.Path;
import sun.misc.Signal;

import java.io.*;
import java.net.*;
import java.util.*;
import java.util.concurrent.locks.LockSupport;
import java.util.zip.ZipEntry;
import java.util.zip.ZipInputStream;

public class ServerMain {
    private static final String VERSION_TXT = "version.txt";

    protected PropServerConfiguration configuration;

    public ServerMain(String[] args) throws Exception {
        //properties fetched from sources other than server.conf
        final BuildInformation buildInformation = BuildInformationHolder.INSTANCE;

        System.err.printf(
                "QuestDB server %s%nCopyright (C) 2014-%d, all rights reserved.%n%n",
                buildInformation.getQuestDbVersion(),
                Dates.getYear(System.currentTimeMillis())
        );
        if (args.length < 1) {
            System.err.println("Root directory name expected");
            return;
        }

        if (Os.type == Os._32Bit) {
            System.err.println("QuestDB requires 64-bit JVM");
            return;
        }

        final CharSequenceObjHashMap<String> optHash = hashArgs(args);

        final Log log = LogFactory.getLog("server-main");
        // expected flags:
        // -d <root dir> = sets root directory
        // -f = forces copy of site to root directory even if site exists
        // -n = disables handling of HUP signal

        final String rootDirectory = optHash.get("-d");
        extractSite(rootDirectory, log);
        final Properties properties = new Properties();
        final String configurationFileName = "/server.conf";
        final File configurationFile = new File(new File(rootDirectory, PropServerConfiguration.CONFIG_DIRECTORY), configurationFileName);

        try (InputStream is = new FileInputStream(configurationFile)) {
            properties.load(is);
        }

        readServerConfiguration(rootDirectory, properties, log, buildInformation);
        log.advisory().$("Server config : ").$(configurationFile.getAbsoluteFile()).$();
        log.advisory().$("Config changes applied:").$();
        log.advisory().$("  http.enabled : ").$(configuration.getHttpServerConfiguration().isEnabled()).$();
        log.advisory().$("  tcp.enabled  : ").$(configuration.getLineTcpReceiverConfiguration().isEnabled()).$();
        log.advisory().$("  pg.enabled   : ").$(configuration.getPGWireConfiguration().isEnabled()).$();

        log.advisory().$("open database [id=").$(configuration.getCairoConfiguration().getDatabaseIdLo()).$('.').$(configuration.getCairoConfiguration().getDatabaseIdHi()).$(']').$();
        log.advisory().$("platform [bit=").$(System.getProperty("sun.arch.data.model")).$(']').$();
        switch (Os.type) {
            case Os.WINDOWS:
                log.advisory().$("OS/Arch: windows/amd64").$(Vect.getSupportedInstructionSetName()).$();
                break;
            case Os.LINUX_AMD64:
                log.advisory().$("OS/Arch: linux/amd64").$(Vect.getSupportedInstructionSetName()).$();
                break;
            case Os.OSX_AMD64:
                log.advisory().$("OS/Arch: apple/amd64").$(Vect.getSupportedInstructionSetName()).$();
                break;
            case Os.OSX_ARM64:
                log.advisory().$("OS/Arch: apple/apple-silicon").$();
                break;
            case Os.LINUX_ARM64:
                log.advisory().$("OS/Arch: linux/arm64").$(Vect.getSupportedInstructionSetName()).$();
                break;
            case Os.FREEBSD:
                log.advisory().$("OS: freebsd/amd64").$(Vect.getSupportedInstructionSetName()).$();
                break;
            default:
                log.critical().$("Unsupported OS").$(Vect.getSupportedInstructionSetName()).$();
                break;
        }
        log.advisory().$("available CPUs: ").$(Runtime.getRuntime().availableProcessors()).$();
        log.advisory().$("db root: ").$(configuration.getCairoConfiguration().getRoot()).$();
        log.advisory().$("backup root: ").$(configuration.getCairoConfiguration().getBackupRoot()).$();
        try (Path path = new Path()) {
            verifyFileSystem("db", configuration.getCairoConfiguration().getRoot(), path, log);
            verifyFileSystem("backup", configuration.getCairoConfiguration().getBackupRoot(), path, log);
        }

        if (JitUtil.isJitSupported()) {
            final int jitMode = configuration.getCairoConfiguration().getSqlJitMode();
            switch (jitMode) {
                case SqlJitMode.JIT_MODE_ENABLED:
                    log.advisory().$("SQL JIT compiler mode: on").$();
                    log.advisory().$("Note: JIT compiler mode is a beta feature.").$();
                    break;
                case SqlJitMode.JIT_MODE_FORCE_SCALAR:
                    log.advisory().$("SQL JIT compiler mode: scalar").$();
                    log.advisory().$("Note: JIT compiler mode is a beta feature.").$();
                    break;
                case SqlJitMode.JIT_MODE_DISABLED:
                    log.advisory().$("SQL JIT compiler mode: off").$();
                    break;
                default:
                    log.error().$("Unknown SQL JIT compiler mode: ").$(jitMode).$();
                    break;
            }
        }

        final WorkerPool workerPool = new WorkerPool(configuration.getWorkerPoolConfiguration());
        final FunctionFactoryCache functionFactoryCache = new FunctionFactoryCache(
                configuration.getCairoConfiguration(),
                ServiceLoader.load(FunctionFactory.class, FunctionFactory.class.getClassLoader())
        );
        final ObjList<Closeable> instancesToClean = new ObjList<>();

        LogFactory.configureFromSystemProperties(workerPool);
        final CairoEngine cairoEngine = new CairoEngine(configuration.getCairoConfiguration());
        workerPool.assign(cairoEngine.getEngineMaintenanceJob());
        instancesToClean.add(cairoEngine);

        if (!configuration.getCairoConfiguration().getTelemetryConfiguration().getDisableCompletely()) {
            final TelemetryJob telemetryJob = new TelemetryJob(cairoEngine, functionFactoryCache);
            instancesToClean.add(telemetryJob);

            if (configuration.getCairoConfiguration().getTelemetryConfiguration().getEnabled()) {
                workerPool.assign(telemetryJob);
            }
        }

        workerPool.assignCleaner(Path.CLEANER);
        O3Utils.setupWorkerPool(workerPool, cairoEngine.getMessageBus());

        Metrics metrics;
        if (configuration.getMetricsConfiguration().isEnabled()) {
            metrics = Metrics.enabled();
        } else {
            metrics = Metrics.disabled();
        }

        try {
            initQuestDb(workerPool, cairoEngine, log);

            instancesToClean.add(createHttpServer(workerPool, log, cairoEngine, functionFactoryCache, metrics));
            instancesToClean.add(createMinHttpServer(workerPool, log, cairoEngine, functionFactoryCache, metrics));

            if (configuration.getPGWireConfiguration().isEnabled()) {
                instancesToClean.add(PGWireServer.create(
                        configuration.getPGWireConfiguration(),
                        workerPool,
                        log,
                        cairoEngine,
                        functionFactoryCache,
                        metrics
                ));
            }

            if (configuration.getLineUdpReceiverConfiguration().isEnabled()) {
                if (Os.type == Os.LINUX_AMD64 || Os.type == Os.LINUX_ARM64) {
                    instancesToClean.add(new LinuxMMLineUdpReceiver(
                            configuration.getLineUdpReceiverConfiguration(),
                            cairoEngine,
                            workerPool
                    ));
                } else {
                    instancesToClean.add(new LineUdpReceiver(
                            configuration.getLineUdpReceiverConfiguration(),
                            cairoEngine,
                            workerPool
                    ));
                }
            }

            instancesToClean.add(LineTcpReceiver.create(
                    configuration.getLineTcpReceiverConfiguration(),
                    workerPool,
                    log,
                    cairoEngine
            ));

            startQuestDb(workerPool, cairoEngine, log);
            if (configuration.getHttpServerConfiguration().isEnabled()) {
                logWebConsoleUrls(log, configuration);
            }

            System.gc();

            log.advisory().$("enjoy").$();

            if (Os.type != Os.WINDOWS && optHash.get("-n") == null) {
                // suppress HUP signal
                Signal.handle(new Signal("HUP"), signal -> {
                });
            }

            Runtime.getRuntime().addShutdownHook(new Thread(() -> {
                System.err.println(new Date() + " QuestDB is shutting down");
                shutdownQuestDb(workerPool, instancesToClean);
                System.err.println(new Date() + " QuestDB is down");
            }));
        } catch (NetworkError e) {
            log.error().$((Sinkable) e).$();
            LockSupport.parkNanos(10000000L);
            System.exit(55);
        }
    }

    private void verifyFileSystem(String kind, CharSequence dir, Path path, Log log) {
        if (dir != null) {
            path.of(dir).$();
            // path will contain file system name
            long fsStatus = Files.getFileSystemStatus(path);
            path.seekZ();
            LogRecord rec = log.advisory().$(kind).$(" file system magic: 0x");
            if (fsStatus < 0) {
                rec.$hex(-fsStatus).$(" [").$(path).$("] SUPPORTED").$();
            } else {
                rec.$hex(fsStatus).$(" [").$(path).$("] EXPERIMENTAL").$();
                log.advisory().$("\n\n\n\t\t\t*** SYSTEM IS USING UNSUPPORTED FILE SYSTEM AND COULD BE UNSTABLE ***\n\n").$();
            }
        }
    }

    public static void deleteOrException(File file) {
        if (!file.exists()) {
            return;
        }
        deleteDirContentsOrException(file);

        int retryCount = 3;
        boolean deleted = false;
        while (retryCount > 0 && !(deleted = file.delete())) {
            retryCount--;
            Thread.yield();
        }

        if (!deleted) {
            throw new RuntimeException("Cannot delete file " + file);
        }
    }

    public static void main(String[] args) throws Exception {
        new ServerMain(args);
    }

    private static void logWebConsoleUrls(Log log, PropServerConfiguration configuration) throws SocketException {
        final LogRecord record = log.info().$("web console URL(s):").$('\n').$('\n');
        final int httpBindIP = configuration.getHttpServerConfiguration().getDispatcherConfiguration().getBindIPv4Address();
        final int httpBindPort = configuration.getHttpServerConfiguration().getDispatcherConfiguration().getBindPort();
        if (httpBindIP == 0) {
            Enumeration<NetworkInterface> nets = NetworkInterface.getNetworkInterfaces();
            for (NetworkInterface networkInterface : Collections.list(nets)) {
                Enumeration<InetAddress> inetAddresses = networkInterface.getInetAddresses();
                for (InetAddress inetAddress : Collections.list(inetAddresses)) {
                    if (inetAddress instanceof Inet4Address) {
                        record.$('\t').$("http:/").$(inetAddress).$(':').$(httpBindPort).$('\n');
                    }
                }
            }
            record.$('\n').$();
        } else {
            record.$('\t').$("http://").$ip(httpBindIP).$(':').$(httpBindPort).$('\n').$();
        }
    }

    private static CharSequenceObjHashMap<String> hashArgs(String[] args) {
        CharSequenceObjHashMap<String> optHash = new CharSequenceObjHashMap<>();
        String flag = null;
        for (String s : args) {
            if (s.startsWith("-")) {
                if (flag != null) {
                    optHash.put(flag, "");
                }
                flag = s;
            } else {
                if (flag != null) {
                    optHash.put(flag, s);
                    flag = null;
                } else {
                    System.err.println("Unknown arg: " + s);
                    System.exit(55);
                }
            }
        }

        if (flag != null) {
            optHash.put(flag, "");
        }

        return optHash;
    }


    private static long getPublicVersion(String publicDir) throws IOException {
        File f = new File(publicDir, VERSION_TXT);
        if (f.exists()) {
            try (FileInputStream fis = new FileInputStream(f)) {
                byte[] buf = new byte[128];
                int len = fis.read(buf);
                return Long.parseLong(new String(buf, 0, len));
            }
        }
        return Long.MIN_VALUE;
    }

    private static void setPublicVersion(String publicDir, long version) throws IOException {
        File f = new File(publicDir, VERSION_TXT);
        try (FileOutputStream fos = new FileOutputStream(f)) {
            byte[] buf = Long.toString(version).getBytes();
            fos.write(buf, 0, buf.length);
        }
    }

    private static void extractSite(String dir, Log log) throws IOException {
        final String publicZip = "/io/questdb/site/public.zip";
        final String publicDir = dir + "/public";
        final byte[] buffer = new byte[1024 * 1024];
        URL resource = ServerMain.class.getResource(publicZip);
        long thisVersion = Long.MIN_VALUE;
        if (resource == null) {
            log.error().$("did not find Web Console build at '").$(publicZip).$("'. Proceeding without Web Console checks").$();
        } else {
            thisVersion = resource.openConnection().getLastModified();
        }
        final long oldVersion = getPublicVersion(publicDir);
        if (thisVersion > oldVersion) {
            try (final InputStream is = ServerMain.class.getResourceAsStream(publicZip)) {
                if (is != null) {
                    try (ZipInputStream zip = new ZipInputStream(is)) {
                        ZipEntry ze;
                        while ((ze = zip.getNextEntry()) != null) {
                            final File dest = new File(publicDir, ze.getName());
                            if (!ze.isDirectory()) {
                                copyInputStream(true, buffer, dest, zip, log);
                            }
                            zip.closeEntry();
                        }
                    }
                } else {
                    log.error().$("could not find site [resource=").$(publicZip).$(']').$();
                }
            }
            setPublicVersion(publicDir, thisVersion);
            copyConfResource(dir, false, buffer, "conf/date.formats", log);
            copyConfResource(dir, true, buffer, "conf/mime.types", log);
            copyConfResource(dir, false, buffer, "conf/server.conf", log);
        } else {
            log.info().$("web console is up to date").$();
        }
    }

    private static void copyConfResource(String dir, boolean force, byte[] buffer, String res, Log log) throws IOException {
        File out = new File(dir, res);
        try (InputStream is = ServerMain.class.getResourceAsStream("/io/questdb/site/" + res)) {
            if (is != null) {
                copyInputStream(force, buffer, out, is, log);
            }
        }
    }

    private static void copyInputStream(boolean force, byte[] buffer, File out, InputStream is, Log log) throws IOException {
        final boolean exists = out.exists();
        if (force || !exists) {
            File dir = out.getParentFile();
            if (!dir.exists() && !dir.mkdirs()) {
                log.error().$("could not create directory [path=").$(dir).$(']').$();
                return;
            }
            try (FileOutputStream fos = new FileOutputStream(out)) {
                int n;
                while ((n = is.read(buffer, 0, buffer.length)) > 0) {
                    fos.write(buffer, 0, n);
                }
            }
            log.info().$("extracted [path=").$(out).$(']').$();
            return;
        }
        log.debug().$("skipped [path=").$(out).$(']').$();
    }

    private static void deleteDirContentsOrException(File file) {
        if (!file.exists()) {
            return;
        }
        try {
            if (notSymlink(file)) {
                File[] files = file.listFiles();
                if (files != null) {
                    for (File f : files) {
                        deleteOrException(f);
                    }
                }
            }
        } catch (IOException e) {
            throw new RuntimeException("Cannot delete dir contents: " + file, e);
        }
    }

    private static boolean notSymlink(File file) throws IOException {
        if (file == null) {
            throw new IllegalArgumentException("File must not be null");
        }
        if (File.separatorChar == '\\') {
            return true;
        }

        File fileInCanonicalDir;
        if (file.getParentFile() == null) {
            fileInCanonicalDir = file;
        } else {
            File canonicalDir = file.getParentFile().getCanonicalFile();
            fileInCanonicalDir = new File(canonicalDir, file.getName());
        }

        return fileInCanonicalDir.getCanonicalFile().equals(fileInCanonicalDir.getAbsoluteFile());
    }

    protected static void shutdownQuestDb(final WorkerPool workerPool, final ObjList<? extends Closeable> instancesToClean) {
        workerPool.halt();
        Misc.freeObjList(instancesToClean);
    }

    protected HttpServer createHttpServer(
            final WorkerPool workerPool,
            final Log log,
            final CairoEngine cairoEngine,
            FunctionFactoryCache functionFactoryCache,
            Metrics metrics) {
        return HttpServer.create(
                configuration.getHttpServerConfiguration(),
                workerPool,
                log,
                cairoEngine,
                functionFactoryCache,
                metrics
        );
    }

    protected HttpServer createMinHttpServer(
            final WorkerPool workerPool,
            final Log log,
            final CairoEngine cairoEngine,
            FunctionFactoryCache functionFactoryCache,
            Metrics metrics) {
        return HttpServer.createMin(
                configuration.getHttpMinServerConfiguration(),
                workerPool,
                log,
                cairoEngine,
                functionFactoryCache,
                metrics
        );
    }

    protected void initQuestDb(
            final WorkerPool workerPool,
            final CairoEngine cairoEngine,
            final Log log
    ) {
        // For extension
    }

    protected void readServerConfiguration(final String rootDirectory,
                                           final Properties properties,
                                           Log log,
                                           final BuildInformation buildInformation) throws ServerConfigurationException, JsonException {
        configuration = new PropServerConfiguration(rootDirectory, properties, System.getenv(), log, buildInformation);
    }

    protected void startQuestDb(
            final WorkerPool workerPool,
            final CairoEngine cairoEngine,
            final Log log
    ) {
        workerPool.start(log);
    }
}